#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""Test the usage of frequency plans."""

import os
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from h5py import File
from scipy.interpolate import interp1d

from lisainstrument import Instrument


def _consistent_locking_beatnotes(instru, fplan=None):
    """Check that TPS locking beatnotes are consistent with ``fplan``.

    We check that locking beatnotes are equal up to machine precision.

    Args:
        instru (:class:`lisainstrument.Instrument`): instrument object
        fplan (dict or None): dictionary of locking beatnotes [Hz] or None to read fplan file

    Returns:
        (bool) Whether the locking beatnotes are valid.
    """
    # Read fplan file
    if fplan is None:
        with File(instru.fplan_file, "r") as fplanf:
            with File(instru.orbit_file, "r") as orbitsf:
                t0 = orbitsf.attrs["t0"]
            t = t0 + np.arange(fplanf.attrs["size"]) * fplanf.attrs["dt"]
            if instru.lock_config == "N1-12":
                fplan = {
                    "13": interp1d(t, -fplanf[instru.lock_config]["rfi_12"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "31": interp1d(t, fplanf[instru.lock_config]["isi_31"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "32": interp1d(t, -fplanf[instru.lock_config]["rfi_31"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "21": interp1d(t, fplanf[instru.lock_config]["isi_21"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "23": interp1d(t, fplanf[instru.lock_config]["rfi_23"][:] * 1e6)(
                        instru.physics_t
                    ),
                }
            elif instru.lock_config == "N1-21":
                fplan = {
                    "23": interp1d(t, fplanf[instru.lock_config]["rfi_23"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "32": interp1d(t, fplanf[instru.lock_config]["isi_32"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "31": interp1d(t, fplanf[instru.lock_config]["rfi_31"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "12": interp1d(t, fplanf[instru.lock_config]["isi_12"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "13": interp1d(t, -fplanf[instru.lock_config]["rfi_12"][:] * 1e6)(
                        instru.physics_t
                    ),
                }
            elif instru.lock_config == "N4-12":
                fplan = {
                    "13": interp1d(t, -fplanf[instru.lock_config]["rfi_12"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "31": interp1d(t, fplanf[instru.lock_config]["isi_31"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "32": interp1d(t, -fplanf[instru.lock_config]["rfi_31"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "23": interp1d(t, fplanf[instru.lock_config]["isi_23"][:] * 1e6)(
                        instru.physics_t
                    ),
                    "21": interp1d(t, fplanf[instru.lock_config]["isi_21"][:] * 1e6)(
                        instru.physics_t
                    ),
                }
    # Check locking beatnotes
    if instru.lock_config == "N1-12":
        return np.all(
            [
                np.allclose(instru.tps_rfi_carrier_offsets["13"], fplan["13"]),
                np.allclose(instru.tps_isi_carrier_offsets["31"], fplan["31"]),
                np.allclose(instru.tps_rfi_carrier_offsets["32"], fplan["32"]),
                np.allclose(instru.tps_isi_carrier_offsets["21"], fplan["21"]),
                np.allclose(instru.tps_rfi_carrier_offsets["23"], fplan["23"]),
            ]
        )
    if instru.lock_config == "N1-21":
        return np.all(
            [
                np.allclose(instru.tps_rfi_carrier_offsets["23"], fplan["23"]),
                np.allclose(instru.tps_isi_carrier_offsets["32"], fplan["32"]),
                np.allclose(instru.tps_rfi_carrier_offsets["31"], fplan["31"]),
                np.allclose(instru.tps_isi_carrier_offsets["12"], fplan["12"]),
                np.allclose(instru.tps_rfi_carrier_offsets["13"], fplan["13"]),
            ]
        )
    if instru.lock_config == "N4-12":
        return np.all(
            [
                np.allclose(instru.tps_rfi_carrier_offsets["13"], fplan["13"]),
                np.allclose(instru.tps_isi_carrier_offsets["31"], fplan["31"]),
                np.allclose(instru.tps_rfi_carrier_offsets["32"], fplan["32"]),
                np.allclose(instru.tps_isi_carrier_offsets["23"], fplan["23"]),
                np.allclose(instru.tps_isi_carrier_offsets["21"], fplan["21"]),
            ]
        )
    raise ValueError(f"unsupported lock configuration '{instru.lock_config}'")


def test_static_fplan():
    """Test the default static set of locking beatnotes."""

    # Check fplan initialization
    instru = Instrument(size=100, fplan="static")
    static = {
        "12": 8e6,
        "23": 9e6,
        "31": 10e6,
        "13": -8.2e6,
        "32": -8.5e6,
        "21": -8.7e6,
    }
    for mosa in instru.MOSAS:
        assert instru.fplan[mosa] == static[mosa]

    # Check locking beatnotes
    instru = Instrument(size=100, lock="N1-12", fplan="static")
    instru.simulate()
    instru.write(mode="w")
    assert _consistent_locking_beatnotes(instru, static)
    instru = Instrument(size=100, lock="N1-21", fplan="static")
    instru.simulate()
    instru.write(mode="w")
    assert _consistent_locking_beatnotes(instru, static)
    instru = Instrument(size=100, lock="N4-12", fplan="static")
    instru.simulate()
    instru.write(mode="w")
    assert _consistent_locking_beatnotes(instru, static)


def test_static_fplan_valid_with_all_lock_configs():
    """Check that 'static' fplan is valid for all lock configs.

    We check that all beatnotes are between 5 and 25 MHz (and negative)
    for all locking configurations, using the default static set of locking
    beatnotes.
    """
    for topology in Instrument.LOCK_TOPOLOGIES:
        instru = Instrument(size=100, lock=f"{topology}-12", fplan="static")
        instru.disable_all_noises()
        instru.simulate()
        instru.write(mode="w")

        for mosa in instru.MOSAS:
            assert np.all(5e6 <= np.abs(instru.isi_carriers[mosa]))
            assert np.all(5e6 <= np.abs(instru.isi_usbs[mosa]))
            assert np.all(5e6 <= np.abs(instru.rfi_carriers[mosa]))
            assert np.all(5e6 <= np.abs(instru.rfi_usbs[mosa]))
            assert np.all(np.abs(instru.isi_carriers[mosa]) <= 25e6)
            assert np.all(np.abs(instru.isi_usbs[mosa]) <= 25e6)
            assert np.all(np.abs(instru.rfi_carriers[mosa]) <= 25e6)
            assert np.all(np.abs(instru.rfi_usbs[mosa]) <= 25e6)


def test_constant_equal_fplan():
    """Test a user-defined constant equal fplan."""

    # Check fplan initialization
    instru = Instrument(size=100, fplan=42.0)
    for mosa in instru.MOSAS:
        assert instru.fplan[mosa] == 42.0


def test_constant_unequal_fplan():
    """Test a user-defined constant unequal fplan."""

    # Check fplan initialization
    fplan = {
        "12": 8.1e6,
        "23": 9.2e6,
        "31": 10.3e6,
        "13": 1.4e6,
        "32": -11.6e6,
        "21": -9.5e6,
    }
    instru = Instrument(size=100, lock="N1-12", fplan=fplan)
    for mosa in instru.MOSAS:
        assert instru.fplan[mosa] == fplan[mosa]

    # Check locking beatnotes
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru = Instrument(size=100, lock="N1-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N1-21", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N4-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)


def test_varying_equal_fplan():
    """Test a user-defined time-varying equal fplan."""

    # Check fplan initialization
    fplan = np.random.uniform(5e6, 25e6, size=400)
    instru = Instrument(size=100, fplan=fplan)
    for mosa in instru.MOSAS:
        assert np.all(instru.fplan[mosa] == fplan)

    # Check locking beatnotes
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        fplan = {mosa: fplan for mosa in Instrument.MOSAS}
        instru = Instrument(size=100, lock="N1-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N1-21", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N4-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)


def test_varying_unequal_fplan():
    """Test a user-defined time-varying unequal fplan."""

    # Check fplan initialization
    fplan = {mosa: np.random.uniform(5e6, 25e6, size=400) for mosa in Instrument.MOSAS}
    instru = Instrument(size=100, fplan=fplan)
    for mosa in instru.MOSAS:
        assert np.all(instru.fplan[mosa] == fplan[mosa])

    # Check locking beatnotes
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru = Instrument(size=100, lock="N1-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N1-21", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)
        instru = Instrument(size=100, lock="N4-12", fplan=fplan)
        instru.simulate()
        instru.write(path, mode="w")
        assert _consistent_locking_beatnotes(instru, fplan)


def test_keplerian_fplan_1_1():
    """Test standard Keplerian fplan file v1.1."""

    # Check fplan file with standard lock configs
    for topology in Instrument.LOCK_TOPOLOGIES:
        instru = Instrument(
            size=100,
            lock=f"{topology}-12",
            fplan="tests/data/keplerian-fplan-1-1.h5",
            orbits="tests/data/keplerian-orbits-1-0-2.h5",  # does not match, but okay
        )

    # Should raise an error for non-standard lock config
    with pytest.raises(ValueError):
        Instrument(
            size=100,
            lock="six",
            fplan="tests/data/keplerian-fplan-1-1.h5",
            orbits="tests/data/keplerian-orbits-1-0-2.h5",
        )
    with pytest.raises(ValueError):
        Instrument(size=100, lock="three", fplan="tests/data/keplerian-fplan-1-1.h5")
    with pytest.raises(ValueError):
        lock = {
            "12": "cavity",
            "13": "cavity",
            "21": "distant",
            "31": "distant",
            "23": "adjacent",
            "32": "adjacent",
        }
        Instrument(size=100, lock=lock, fplan="tests/data/keplerian-fplan-1-1.h5")

    # Should raise an error without orbit file
    with pytest.raises(ValueError):
        Instrument(size=100, lock="N1-12", fplan="tests/data/keplerian-fplan-1-1.h5")

    # Check locking beatnotes
    with File("tests/data/keplerian-orbits-1-0-2.h5", "r") as f:
        t0 = f.attrs["t0"] + 100
    instru = Instrument(
        size=100,
        lock="N1-12",
        t0=t0,
        fplan="tests/data/keplerian-fplan-1-1.h5",
        orbits="tests/data/keplerian-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)
    instru = Instrument(
        size=100,
        lock="N1-21",
        t0=t0,
        fplan="tests/data/keplerian-fplan-1-1.h5",
        orbits="tests/data/keplerian-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)
    instru = Instrument(
        size=100,
        lock="N4-12",
        t0=t0,
        fplan="tests/data/keplerian-fplan-1-1.h5",
        orbits="tests/data/keplerian-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)


def test_esa_trailing_fplan_1_1():
    """Test standard ESA trailing fplan file v1.1."""

    # Check fplan file with standard lock configs
    for topology in Instrument.LOCK_TOPOLOGIES:
        instru = Instrument(
            size=100,
            lock=f"{topology}-12",
            fplan="tests/data/esa-trailing-fplan-1-1.h5",
            orbits="tests/data/esa-orbits-1-0-2.h5",  # does not match, but okay
        )

    # Should raise an error for non-standard lock config
    with pytest.raises(ValueError):
        Instrument(
            size=100,
            lock="six",
            fplan="tests/data/esa-trailing-fplan-1-1.h5",
            orbits="tests/data/esa-orbits-1-0-2.h5",
        )
    with pytest.raises(ValueError):
        Instrument(size=100, lock="three", fplan="tests/data/esa-trailing-fplan-1-1.h5")
    with pytest.raises(ValueError):
        lock = {
            "12": "cavity",
            "13": "cavity",
            "21": "distant",
            "31": "distant",
            "23": "adjacent",
            "32": "adjacent",
        }
        Instrument(size=100, lock=lock, fplan="tests/data/esa-trailing-fplan-1-1.h5")

    # Should raise an error without orbit file
    with pytest.raises(ValueError):
        Instrument(size=100, lock="six", fplan="tests/data/esa-trailing-fplan-1-1.h5")

    # Check locking beatnotes
    with File("tests/data/esa-orbits-1-0-2.h5", "r") as f:
        t0 = f.attrs["t0"] + 100
    instru = Instrument(
        size=100,
        lock="N1-12",
        t0=t0,
        fplan="tests/data/esa-trailing-fplan-1-1.h5",
        orbits="tests/data/esa-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)
    instru = Instrument(
        size=100,
        lock="N1-21",
        t0=t0,
        fplan="tests/data/esa-trailing-fplan-1-1.h5",
        orbits="tests/data/esa-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)
    instru = Instrument(
        size=100,
        lock="N4-12",
        t0=t0,
        fplan="tests/data/esa-trailing-fplan-1-1.h5",
        orbits="tests/data/esa-orbits-1-0-2.h5",
    )
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert _consistent_locking_beatnotes(instru)


def test_offset_freqs():
    """Test that offset frequencies are correctly set for unlocked lasers."""

    # Check that default set works for 'six' lock config
    instru = Instrument(size=100, lock="six", offset_freqs="default")
    instru.disable_all_noises()
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    for mosa in instru.MOSAS:
        assert np.all(5e6 <= np.abs(instru.isi_carriers[mosa]))
        assert np.all(5e6 <= np.abs(instru.isi_usbs[mosa]))
        assert np.all(5e6 <= np.abs(instru.rfi_carriers[mosa]))
        assert np.all(5e6 <= np.abs(instru.rfi_usbs[mosa]))
        assert np.all(np.abs(instru.isi_carriers[mosa]) <= 25e6)
        assert np.all(np.abs(instru.isi_usbs[mosa]) <= 25e6)
        assert np.all(np.abs(instru.rfi_carriers[mosa]) <= 25e6)
        assert np.all(np.abs(instru.rfi_usbs[mosa]) <= 25e6)

    # Check that unlocked laser frequency offsets are correctly set for 'six' lock config
    offset_freqs = {
        "12": 8.1e6,
        "23": 9.2e6,
        "31": 10.3e6,
        "13": 1.4e6,
        "32": -11.6e6,
        "21": -9.5e6,
    }
    instru = Instrument(size=100, lock="six", offset_freqs=offset_freqs)
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    for mosa in instru.MOSAS:
        assert instru.local_carrier_offsets[mosa] == offset_freqs[mosa]

    # Check that unlocked laser frequency offsets are correctly set for 'three' lock config
    offset_freqs = {
        "12": 8.1e6,
        "23": 9.2e6,
        "31": 10.3e6,
        "13": 1.4e6,
        "32": -11.6e6,
        "21": -9.5e6,
    }
    fplan = {
        "12": 8e6,
        "23": 9e6,
        "31": 10e6,
        "13": -8.2e6,
        "32": -8.5e6,
        "21": -8.7e6,
    }
    instru = Instrument(
        size=100,
        lock="three",
        offset_freqs=offset_freqs,
        clock_freqoffsets=0,
        clock_freqlindrifts=0,
        clock_freqquaddrifts=0,
    )
    instru.write(keep_all=True, mode="w")
    with File("measurements.h5", "r") as hdf5:
        local_carrier_offsets = hdf5["local_carrier_offsets"][()]
    for mosa in instru.MOSAS[:3]:
        assert local_carrier_offsets[mosa] == offset_freqs[mosa]
    for left, right in zip(["12", "23", "31"], ["13", "21", "32"]):
        assert local_carrier_offsets[right] == offset_freqs[left] - fplan[right]

    # Check that frequency offsets are only set on unlocked lasers
    instru = Instrument(size=100, lock="N1-12", offset_freqs=offset_freqs)
    instru.disable_all_noises()
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert instru.local_carrier_offsets["12"] == offset_freqs["12"]
    instru = Instrument(size=100, lock="N1-23", offset_freqs=offset_freqs)
    instru.disable_all_noises()
    instru.simulate()
    with TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, "measurements.h5")
        instru.write(path, mode="w")
    assert instru.local_carrier_offsets["23"] == offset_freqs["23"]
